require 'rails_helper'
# %w[
#   applicant
#   housing_request
# ].each { |factory_name|require Hermes::Engine.root.join('spec', 'factories', "#{factory_name.pluralize}_factory").to_s }

RSpec.describe Hermes::HousingRequest, type: :model do
  describe 'basic controls' do
    let!(:user1) { create(:user) }
    let!(:applicant1) do
      create(
        :applicant,
        user: user1,
        monthly_charges: {}
      )
    end

    it 'has a valid factory' do
      expect(FactoryBot.build(:housing_request, main_applicant: applicant1)).to be_valid
    end
  end

  context 'with simple setup' do
    let!(:user1) { create(:user) }
    let!(:user2) { create(:user) }
    let!(:user3) { create(:user) }
    let!(:applicant1) {
      create(
        :applicant,
        user: user1,
        monthly_charges: {}
      )
    }
    let!(:applicant2) {
      create(
        :applicant,
        user: user2
      )
    }
    let!(:applicant3) {
      create(
        :applicant,
        user: user3
      )
    }

    let!(:housing_request1) do
      create(
        :housing_request,
        is_active: true,
        main_applicant: applicant1
      )
    end

    let!(:housing_request2) do
      create(
        :housing_request,
        is_active: true,
        main_applicant: applicant1,
        coapplicants: [
          { rel_type: 'Conjoint (e)', id: applicant2.id },
          { rel_type: 'Parent (e)', id: applicant3.id }
        ]
      )
    end

    describe 'Eligibility query are valide' do
      it 'should include right typology ranges' do
        typology_range1 = housing_request1.calculate_typology_range
        typology_range2 = housing_request2.calculate_typology_range
        range1_to_i = typology_range1.map { |typology| Integer(typology.sub('T', '')) }
        range2_to_i = typology_range2.map { |typology| Integer(typology.sub('T', '')) }

        (1..2).each do |n|
          (1..2).each do |i|
            expect(binding.local_variable_get("range#{n}_to_i".to_sym)).to include i
          end
        end
        expect(range1_to_i.max).to be 2
        expect(range2_to_i).to include 3
        expect(range2_to_i.max).to be 3
      end

      it 'eligibility should incorpore typology' do
        typology_filter1 = housing_request1.typology_filter
        expect(typology_filter1.first[:$or][0][:typology]).to eq 'T1'
        expect(typology_filter1.first[:$or][1][:typology]).to eq 'T2'
      end
      describe 'edit action dont duplicate coapplicants' do
       
        before :each do
          housing_request2.coapplicants = [
              { rel_type: 'Conjoint (e)', id: applicant3.id }
          ]
          housing_request2.seeking_elevator = false
          housing_request2.save
          housing_request2.reload
          housing_request2.coapplicants = [{ rel_type: 'M', id: applicant3.id }]
          applicant3.destroy
          ap housing_request2.coapplicants
          puts applicant3.reload.id.to_s + " " + applicant2.id.to_s
          housing_request2.coapplicants
          housing_request2.save
          housing_request2.reload
          housing_request2.forcing_duplicate = true
          housing_request2.save
          housing_request2.reload
        end
        it 'coapplicants doesnt change' do
          ap housing_request2.reload.applicant_ids
          ap housing_request2.reload.coapplicants
          
          expect(housing_request2.reload.applicant_ids.length).to eq(0)
        end
      end
    end

    describe 'Relashinships are valide' do
      it 'should has correct housing_requests situation' do
        expect(applicant1.first_active_request.id).to eq(housing_request2.id)
        expect(applicant1.housing_requests.length).to eq(2)
        expect(applicant1.housing_requests.actives.length).to eq(1)
      end

      it 'should has correct values' do
        revenues_sum = (applicant1.monthly_revenues || 0) + (applicant2.monthly_revenues || 0)
        expect(housing_request2.monthly_revenues).to eq(revenues_sum)
      end

      it 'should has correct coapplicants' do
        coapplicants = housing_request2.coapplicants
        expect(coapplicants.length).to eq(2)
        expect(coapplicants.map { |coapplicant| coapplicant[:rel_type] }).to include('Conjoint (e)').and include('Parent (e)')
        expect(coapplicants.map { |coapplicant| coapplicant[:id] }).to include(applicant2.id).and include(applicant3.id)
      end

      describe 'after coapplicants update' do
        before :each do
          housing_request2.coapplicants = [
            { rel_type: 'Conjoint (e)', id: applicant3.id }
          ]
          housing_request2.save
        end

        it 'should has updated with valid data' do
          r_request = housing_request2.reload
          expect(r_request.applicant_ids.length).to be 1
          expect(r_request.applicant_ids.first).to eq applicant3.id
          expect(r_request.coapplicants.first[:rel_type]).to eq 'Conjoint (e)'
          expect(Hermes::ApplicantLink.where(housing_request_id: r_request.id).count).to be 1
        end
      end
    end

    describe 'applicant should have only public offer as eligibility when last active request becomes inactive' do
      it 'should has first active request with valid data' do
        expect(housing_request2.main_applicant.housing_requests.actives.count).to eq 1
        expect(housing_request2.is_active).to be true
        expect(housing_request2.main_applicant.first_active_request.id.to_s).to eq housing_request2.id.to_s
      end

      context 'is_active is set to false' do
        before :each do
          housing_request2.update(is_active: false)
        end

        it 'should has is_active == false' do
          expect(housing_request2.reload.is_active).to be false
        end

        it 'should has its main_applicant.eligibility == {is_public: true}' do
          expect(housing_request2.reload.main_applicant.eligibility_query).to eq '{"is_public":true}'
        end
      end
    end
    
  end

  # TODO: test is valid but RSpec can't handle exceptions
  #   using rspec-expectations could be a solution
  #
  # context 'with low quantity' do
  #   2.times do |i|
  #     let!("applicant#{i}") { create(:applicant) }
  #     let!("request#{i}") { create(:housing_request, main_applicant: send("applicant#{i}")) }
  #   end

  #   describe 'set als_id with same data' do
  #     before :each do
  #       hex = '1234567890abcdef01'
  #       2.times do |i|
  #         target_offer = send("request#{i}")
  #         target_offer.als_id = hex
  #         case i
  #         when 0
  #           expect(target_offer.save).to be true
  #         when 1
  #           begin
  #             target_offer.save
  #           rescue ArgumentError => e
  #             ap e
  #             expect(e.message).to eq "#{hex} already exists for als_id"
  #           end
  #         end
  #       end
  #     end

  #     it 'should has the same als_id' do
  #       als_ids = Hermes::HousingRequest.only(:als_id).pluck(:als_id)
  #       puts als_ids.to_s
  #       expect(als_ids.each_cons(2).all? { |a, b| a != b }).to be true
  #     end
  #   end
  # end

  context 'with large quantity' do
    nb = 20
    nb.times do |i|
      let!("applicant#{i}") { create(:applicant) }
      let!("request#{i}") { create(:housing_request, main_applicant: send("applicant#{i}")) }
    end

    describe 'als_id is set' do
      let(:all_requests) { Hermes::HousingRequest.only(:als_id).where(als_id: { :$ne => nil }) }

      it "has #{nb} entities" do
        expect(all_requests.count).to be nb
      end

      let(:als_ids) { all_requests.pluck(:als_id).uniq.compact }

      it 'produces valid lenght' do
        expect(als_ids.map(&:length)).to all(be(18))
      end

      it 'has no duplicates' do
        expect(
          als_ids.map { |als_id| all_requests.where(als_id: als_id).count }
        ).to all(be(1))
      end
    end
  end

  context 'different applicant profil' do
    let!(:user) { create(:user) }
    let!(:user2) { create(:user) }

    describe 'category types as expected' do
      
      it 'Profils 7 : 1 pers avec 1 enfant à charge = catégorie de ménage 3' do
        applicant = create(
          :applicant,
          user: user,
          family_situation: 'Célibataire',
          monthly_charges: {}
        )
        housing_request = create(
          :housing_request,
          is_active: true,
          main_applicant: applicant,
          dependents: [
            {
              rel_type: 'enfant à charge',
              birth_date: '2017-03-23',
              first_name: 'Paul',
              last_name: 'Dupont'
            },
          ]
        )

        expect(housing_request.reload.category_type).to be 3
      end

      it 'Profils 8 : 1 pers avec 1 enfant en droit de visite = catégorie de ménage 2' do
        applicant = create(
          :applicant,
          user: user,
          family_situation: 'Célibataire',
          monthly_charges: {}
        )
        housing_request = create(
          :housing_request,
          is_active: true,
          main_applicant: applicant,
          dependents: [
            {
              rel_type: 'droit de visite',
              birth_date: '2017-03-23',
              first_name: 'Paul',
              last_name: 'Dupont'
            },
          ]
        )

        expect(housing_request.reload.category_type).to be 2
      end

      it 'Profil 9 : 2 pers avec 1 enfant en droit de visite = catégorie de ménage 3' do
        applicant = create(
          :applicant,
          user: user,
          monthly_charges: {}
        )
        coapp = create(
          :applicant,
          user: user2,
          monthly_charges: {}
        )
        housing_request = create(
          :housing_request,
          is_active: true,
          main_applicant: applicant,
          coapplicants: [
            { rel_type: 'Marié (e)', id: coapp.id }
          ],
          dependents: [
            {
              rel_type: 'droit de visite',
              birth_date: '2017-03-23',
              first_name: 'Paul',
              last_name: 'Dupont'
            },
          ]
        )

        expect(housing_request.reload.category_type).to be 3
      end

      it 'Profils 10 : 1 pers avec 1 enfant en garde alternée = catégorie de ménage 3' do
        applicant = create(
          :applicant,
          user: user,
          family_situation: 'Célibataire',
          monthly_charges: {}
        )
        housing_request = create(
          :housing_request,
          is_active: true,
          main_applicant: applicant,
          dependents: [
            {
              rel_type: 'garde alternée',
              birth_date: '2017-03-23',
              first_name: 'Paul',
              last_name: 'Dupont'
            },
          ]
        )
        
        expect(housing_request.reload.category_type).to be 3
      end

      it 'Profil 11 : 2 pers avec 1 enfant à charge = catégorie de ménage 3' do
        applicant = create(
          :applicant,
          user: user,
          monthly_charges: {}
        )
        coapp = create(
          :applicant,
          user: user2,
          monthly_charges: {}
        )
        housing_request = create(
          :housing_request,
          is_active: true,
          main_applicant: applicant,
          coapplicants: [
            { rel_type: 'Marié (e)', id: coapp.id }
          ],
          dependents: [
            {
              rel_type: 'enfant à charge',
              birth_date: '2017-03-23',
              first_name: 'Paul',
              last_name: 'Dupont'
            },
          ]
        )

        expect(housing_request.reload.category_type).to be 3
      end

    end
  end
end
